/*
 *  Copyright 2008 Ndapandula Nakashole
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileReader;
import java.io.FileWriter;

/**
 * Convenince command-line class to split a directory recursively into smaller 
 * subsets and to hash the each file into a server space. Try <code>HashFiles -h</code> for possible options.
 *
 * @author Edgar Meij
 */
public class HashFiles {
    
    private static String inputDir      = null;
    private static String outputDir     = null;
    private static int limit            = -1;
    private static int outputCounter    = 1;
    private static int fileCounter      = 0;
    private static int server           = 0;
    private static int servers          = 0;
    private static long outputSize      = 0;
    private static File currentOutput   = null;
    private static File output          = null;
    
    private static boolean overWrite    = false;
    private static boolean useSize      = false;
    private static boolean useNumber    = false;
    private static boolean verbose      = false;
 
    
    private static String root = null;
    
    public static void main(String[] args) throws Exception{
        
        if (args.length == 0)
            printUsage();
        
        for (int i=0; i<args.length; ++i)  {
            if (args[i].equals("-h")) 
                printUsage();
            else if (args[i].equals("-input")) {
                i++;
                if (i>=args.length)
                    throw new IllegalArgumentException("specify source dir after -input");
                inputDir = args[i];
            } else if (args[i].equals("-output")) {
                i++;
                if (i>=args.length)
                    throw new IllegalArgumentException("specify destination output dir after -output");
                outputDir = args[i];
            } else if (args[i].equals("-max")) {
                i++;
                if (i>=args.length)
                        throw new IllegalArgumentException("specify split threshold size after -max");
                try {
                    limit = Integer.parseInt(args[i]);
                } catch (java.lang.NumberFormatException ex) { 
                    throw new IllegalArgumentException("specify proper number after -max");
                }
            } else if (args[i].equals("-servers")) {
                i++;
                if (i>=args.length)
                        throw new IllegalArgumentException("specify number of servers after -servers");
                try {
                    servers = Integer.parseInt(args[i]);
                } catch (java.lang.NumberFormatException ex) { 
                    throw new IllegalArgumentException("specify proper number after -servers");
                }
            }else if (args[i].equals("-type")) {
                i++;
                if (i>=args.length)
                        printUsage();
                if (args[i].equalsIgnoreCase("size"))
                    useSize=true;
                else if (args[i].equalsIgnoreCase("number"))
                    useNumber=true;
                else 
                    printUsage();
            } else if (args[i].equals("-F")) {
                overWrite = true;
            } else if (args[i].equals("-v")) {
                verbose = true;                
            } else 
                throw new IllegalArgumentException("Unknown argument: " + args[i]);
        }
        
        if (inputDir == null || outputDir == null)
            printUsage();   
        
        if (!useSize && !useNumber) 
            printUsage();
        
        if (useSize)                
            limit *= 1048576; // We want MB's
        
        output = new File(outputDir);
        
        if (output.mkdir()) {
            if (verbose)
                System.err.println(" - '" + output.toString() + "' succesfully created");
            for(int j=1;j<servers;j++){
                File out = new File(output,Integer.toString(j));
                out.mkdir();
            }
        } else {
            if (verbose)
                System.err.println(" - '" + output.toString() + "' exists");
            if (overWrite) {
                if (verbose)
                    System.err.println(" - Deleting all files");
                deleteAllFiles(output);
                output.mkdir();
            }
            else
                System.exit(-1);
        }

        File input = new File(inputDir);
        root = input.getAbsolutePath();

        if(!input.exists()) {
            System.err.println(" - '" + input.toString() + "' does not exist");
            System.exit(-1);
        }

        processInput(input);

        System.out.println();
        System.out.println( fileCounter + " files processed,");
        System.out.println( outputCounter + " directories created.");

    }

    public static void processInput(File dir)  throws Exception{

        if (verbose)
            System.err.println(" - Processing: " + dir.getName() + ", length: " + (dir.length()/1024) + "KB");

        if(dir.isDirectory()) {
            File[] files = dir.listFiles(); 
            for(int i = 0; i < files.length; i++) 
                processInput(files[i]);
        } else {

            long currentSize = 0;
            fileCounter++;

            if (useSize)
                currentSize = dir.length();
            else // useNumber
                currentSize = 1;

            if (!(outputSize + currentSize > limit)) { //max-size limit will be exceeded
//                 outputCounter++;
//                 outputSize = 0;
                server = dir.hashCode();
                System.out.println("hashcode: "+server+ "\tServer: "+server%servers);
//                 currentOutput = new File(output, Integer.toString(server)); //hash file to server space
//                 if (verbose)
//                     System.err.println("   - Creating new dir: " + outputCounter);
// 
//                 outputSize += currentSize;
//                 String parent = getdirs(dir.getName(), dir.getAbsolutePath(),root, currentOutput);
//                 if(parent.equals(null)){
//                     copy(dir, new File(currentOutput, dir.getName()));
//                 }
//                 else{
//                     File prnt =new File(parent);
//                     prnt.mkdirs();
//                     copy(dir, new File(prnt, dir.getName()));
//                 } 
           }else{       //we have processed enough files
                return;
           }

        }
    }
    
    public static void copy(File from, File to) {
        
        try {
            FileReader in = new FileReader(from);
            FileWriter out = new FileWriter(to);
            int c;

            while ((c = in.read()) != -1)
                out.write(c);

            in.close();
            out.close();
        } catch (FileNotFoundException ex) {
            ex.printStackTrace();
            System.exit(-1);
        } catch (IOException ex) {
            ex.printStackTrace();
            System.exit(-1);
        }
    }    
    
    public static boolean deleteAllFiles(File dir) { 
        if(!dir.exists()) 
            return true; 
        
        boolean res = true; 
        
        if(dir.isDirectory()) {
            File[] files = dir.listFiles(); 
            for(int i = 0; i < files.length; i++) 
                res &= deleteAllFiles(files[i]);
            res = dir.delete();//Delete dir itself
        } else 
            res = dir.delete(); 
        
        return res;
    }    
    
    public static void printUsage() {
        String usage = "\nUsage: \njava " + HashFiles.class.getName() +
                " -input SRC_DIR -output DEST_DIR -type TYPE [OPTION]\n\n" +
                "Options\n" +
                " -type TYPE\tset split condition on either:\n" +
                " \t\t size\n" +
                " \t\t number\n" +
                " -max SIZE\tset split threshold (in nr. of files or in MB)\n" +
                " -servers S\tset total number of servers\n" +
                " -F\t\tforce overwriting of files in DEST_DIR\n" +
                " -v\t\tverbose output\n";
        System.err.println(usage);
        
        System.exit(-1);
    } 

        /** 
     * Gets all the necessary dirs containing the file specified by the path
     *
     * @param name  The actual name of the file
     * @param path  The absolute path of the file
     * @param root  The absolute path of the directory containing the file
     */
    public static String getdirs(String name, String path, String root, File dest)throws Exception{
        //strip out all words before and including the root directory name
        int start = path.indexOf(root) + root.length();
        int end = path.lastIndexOf(name);
        String nm = path.substring(start, end);
        System.out.println("name: " +nm);
        nm =dest.getAbsolutePath().concat(nm);
        return nm;
    }
}