/*
 *  Copyright 2006 Edgar Meij
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package indexjob;
import java.io.IOException;

import edu.sdsc.grid.io.MetaDataTable;
import edu.sdsc.grid.io.srb.SRBMetaDataSet;
import edu.sdsc.grid.io.MetaDataCondition;
import edu.sdsc.grid.io.MetaDataSet;
import edu.sdsc.grid.io.MetaDataSelect;
import edu.sdsc.grid.io.MetaDataRecordList;
import edu.sdsc.grid.io.MetaDataTable;
import edu.sdsc.grid.io.srb.SRBMetaDataRecordList;
import edu.sdsc.grid.io.srb.SRBException;
import edu.sdsc.grid.io.srb.SRBFile;

import edu.sdsc.grid.io.srb.SRBFileSystem;
import edu.sdsc.grid.io.local.LocalFile;
import edu.sdsc.grid.io.srb.SRBAccount;
import edu.sdsc.grid.io.MetaDataRecordList;

/**
 * Starting point for all GridLucene SRB methods/activities.
 *
 * @author edgar
 */
public class SRBHandler {
    
    private SRBAccount srbAccount = null;
    private SRBFileSystem srbFileSystem = null;
    private MetaDataRecordList[] rl = null;
    
    /**
     * Creates a new instance of SRBHandler, using (hardcoded) default settings, or 
     * the <code>~/.srb/*</code> files
     */
    public SRBHandler() 
            throws IOException {
        
        // You can put default settings here:
        //this(username, password, port, SRB server, homedir, domain, resource);        
		SRBAccount.setVersion(SRBAccount.SRB_VERSION_3_4);
        srbAccount = new SRBAccount();
        srbFileSystem = new SRBFileSystem( srbAccount );
		if (srbFileSystem == null) {
	        throw new IOException("bla");
		}
	}
    
    /**
     * Creates a new instance of SRBHandler, using credentials
     */
    public SRBHandler(String username, String password, int port, String hostname, String homedir, String domain, String resource) 
            throws IOException {
        
        SRBAccount.setVersion(SRBAccount.SRB_VERSION_3_4);
        srbAccount = new SRBAccount(hostname, port, username, password, homedir, domain, resource);
        srbFileSystem = new SRBFileSystem( srbAccount );    
        if (!srbFileSystem.isConnected()) 
            throw new IOException("Could not connect to filesystem");
        
    }    
    
    
    /**
     * Creates a new instance of SRBHandler, using credentials
     */
    public SRBHandler(LocalFile Envfile, LocalFile Authfile) 
            throws IOException {
        
        SRBAccount.setVersion(SRBAccount.SRB_VERSION_3_4);
        srbAccount = new SRBAccount(Envfile, Authfile);
        srbFileSystem = new SRBFileSystem( srbAccount );    
        if (!srbFileSystem.isConnected()) 
            throw new IOException("Could not connect to filesystem");
        
    }  
    /**
     * Creates a new instance of SRBHandler, using credentials and some less defaults
     */
    public SRBHandler(String username, String password) 
            throws IOException {
        
        // Put default values here:
        //this(username, password, port, SRB server, homedir, domain, resource);        
    }        
    
    /**
     * Returns the SRBaccount 
     * 
     */        
    public SRBAccount getSRBAccount () {
        return srbAccount;
    }
    
    /**
     * Returns the SRBFileSystem 
     *
     */        
    public SRBFileSystem getSRBFileSystem () {
        return srbFileSystem;
    }
    
    /**
     * Changes the SRB metadata for a given dir, using the new value for 
     * the specified key. Adds metadata if it didn't exist already.
     * 
     * 
     * @param dir  The name of the directory, relative to the homeDir of which to change the metadata.
     * @param key  The key name for which to change the metadata
     * @param value The new value for the specified key
     */        
    public void changeDirectoryMetaData(String dir, String key, String value) 
        throws IOException {
        
        boolean done = false;
        SRBFile file = new SRBFile(srbFileSystem, dir);
        
        MetaDataCondition conditions[] = {
            MetaDataSet.newCondition( SRBMetaDataSet.DIRECTORY_NAME,
                MetaDataCondition.EQUAL, 
                srbFileSystem.getHomeDirectory().concat(SRBFile.PATH_SEPARATOR).concat(dir))
        };
        
        String[] selectFieldNames = {
                SRBMetaDataSet.DIRECTORY_NAME,
                SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES
        };
        
        MetaDataSelect selects[] =
                MetaDataSet.newSelection( selectFieldNames );        
        
        MetaDataRecordList[] MDrecList = srbFileSystem.query(conditions, selects, 1);
        
        if (MDrecList == null) {
            throw new IOException("No dir found with the name: " + 
                srbFileSystem.getHomeDirectory().concat(SRBFile.PATH_SEPARATOR).concat(dir));
        } else { 
            //System.out.print(MDrecList.length + "\n");
            for (int i =  0 ; i < MDrecList.length ; i++) {
                

                MetaDataTable out;
                try {
                    out = (MetaDataTable) MDrecList[i].getValue(SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES);

                    MDrecList[i].removeRecord(SRBMetaDataSet.getField(SRBMetaDataSet.DIRECTORY_NAME));                                

                    for (int row=0 ; row < out.getRowCount() ; row++) {

                        if (out.getStringValue(row, 0).equals(key)) {
                            
                            MDrecList[i].removeRecord(SRBMetaDataSet.getField(SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES));
                            MDrecList[i].addRecord(SRBMetaDataSet.getField(SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES), out);
                            out.setStringValue(row, 1, value);
                            
                            deleteDirectoryMetaData(dir);
                            file.modifyMetaData(MDrecList[i]);
                            done = true;
                        }                        
                    }
                } catch (NullPointerException ex) {
                    
                }
            }
            
            if (!done) {
                String[][] definableMetaDataValues = new String[1][2];
                definableMetaDataValues[0][0] = key;
                definableMetaDataValues[0][1] = value;             
            
                addDirectoryMetaData(dir, definableMetaDataValues);                
                //addDirectoryMetaData(dir, new String[][] {{key},{value}});
            }
        }
    }

    /**
     * Deletes the SRB metadata for a given dir.
     * 
     * @param dir  The name of the directory, relative to the homeDir of which to delete the metadata.
     */        
    public void deleteDirectoryMetaData (String dir) 
        throws IOException {
        
        SRBFile file = new SRBFile(srbFileSystem, dir);
        
        MetaDataRecordList empty = new 
            SRBMetaDataRecordList( SRBMetaDataSet.getField(
            SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES ), 
            (MetaDataTable)null );                            

        file.modifyMetaData(empty);
    }
    
    /**
     * Adds new SRB metadata for a given dir, using key-value pairs.
     * 
     * @param dir  The name of the directory, relative to the homeDir of which to change the metadata.
     * @param definableMetaDataValues  An two-dimensional array, containing the metadata values
     */        
    public void addDirectoryMetaData(String dir, String[][] definableMetaDataValues) 
        throws IOException {
        
        SRBFile file = new SRBFile(srbFileSystem, dir);
        
        if (!file.exists())
            throw new IOException(dir + " does not exist");        
        if (!file.isDirectory())
            throw new IOException(dir + " is not a directory");
        if (!file.canRead() || !file.canWrite())
            throw new IOException("Cannot open: " + dir);
        
        int[] operators = new int[definableMetaDataValues.length];
        for (int i = 0 ; i < definableMetaDataValues.length ; i++) {
            operators[i] = MetaDataCondition.EQUAL;
        }
        
        MetaDataTable metaDataTable = new MetaDataTable(operators,
           definableMetaDataValues);        

        SRBMetaDataRecordList recordList = new SRBMetaDataRecordList( SRBMetaDataSet.getField( SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES ), metaDataTable );

        file.modifyMetaData( recordList );
    }

    /**
     * Returns an array containing the locations of directories, matching a given 
     * metadata query
     * 
     * @param definableMetaDataValues  An two-dimensional array, containing the metadata values
     * @return an array containing matching directory names
     */            
    public String[] getMatchingDirectories(String[][] definableMetaDataValues) 
            throws IOException, SRBException{
        
        String[] results; 
        int totalHits;

        int[] operators = new int[definableMetaDataValues.length];
        for (int i = 0 ; i < definableMetaDataValues.length ; i++) {
            operators[i] = MetaDataCondition.EQUAL;
        }
        
        
        MetaDataTable metaDataTableQ = new MetaDataTable(operators,
                                           definableMetaDataValues);
        /**
         * The SRB supports user definable metadata. Access to this
         * data is provided through the MetaDataTable class.
         */
        MetaDataCondition[] conditionsQ = new MetaDataCondition[1];

        conditionsQ[0] = MetaDataSet.newCondition(
            SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES, metaDataTableQ);

        /**
         * For items which matched the query, met the conditions above,
         * the following values will be returned.
         */
        String[] selectFieldNamesQ = {
            SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES
        };
        
        MetaDataSelect selectsQ[] =
            MetaDataSet.newSelection(selectFieldNamesQ);

        selectsQ[0] = MetaDataSet.newSelection(
            SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES);
        rl = srbFileSystem.query(conditionsQ, selectsQ);
        
        if (rl != null) { // Something in the database matched the query
            // For some reason, SRB returns double results
            totalHits = rl.length;
            results = new String[totalHits/2];            
            int counter = 0;
            for (int i = totalHits / 2; i < totalHits; i++) {
                /*
                // System.out.println("i=" + i + ", value: " + rl[i]);
                System.out.println(
                    "Absolute dir name: '"
                    + rl[i].getValue(SRBMetaDataSet.DIRECTORY_NAME) + "'");
                */
                results[counter] = rl[i].getValue(
                        SRBMetaDataSet.DIRECTORY_NAME).toString().replaceAll(
                        srbFileSystem.getHomeDirectory() + "/", "");
                counter++;
            }
            return results;
        } else {
            return null;
        }  
    }
}
