/*
 *  Copyright 2006 Edgar Meij
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.apache.lucene.store;

import java.io.File;
import java.io.IOException;

import java.net.URI;
import java.lang.NullPointerException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import indexjob.SRBHandler;

import edu.sdsc.grid.io.srb.SRBAccount;
import edu.sdsc.grid.io.srb.SRBFileSystem;
import edu.sdsc.grid.io.srb.SRBFile;
import edu.sdsc.grid.io.srb.SRBContainer;
import edu.sdsc.grid.io.srb.SRBException;
import edu.sdsc.grid.io.srb.SRBRandomAccessFile;
import edu.sdsc.grid.io.local.LocalFile;
import edu.sdsc.grid.io.local.LocalRandomAccessFile;
import edu.sdsc.grid.io.srb.SRBFileInputStream;
import edu.sdsc.grid.io.srb.SRBFileOutputStream;

//import org.apache.lucene.store.*;
/*
* Implementation of a Lucene Directory class on SRB. 
* Opens a Lucene index stored on SRB.
* 
* @author: Edgar Meij
*
*/ 
public class SRBDirectory extends Directory {

    private SRBAccount srbAccount = null;
    private SRBFileSystem srbFileSystem = null;    
    private SRBFile srbFile = null;
    private SRBContainer srbContainer = null;
    private SRBRandomAccessFile srbRandomAccessFile = null;
    private SRBFile lockDir; 
    private static MessageDigest DIGESTER;    

    static {
        try {
            DIGESTER = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e.toString());
        }
    }
    
  /** A buffer optionally used in renameTo method */
  private byte[] buffer = null;    
  
  /**
   * So we can do some byte-to-hexchar conversion below
   */
  private static final char[] HEX_DIGITS =
  {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};    
    
    public SRBDirectory (SRBFile indexDir) 
        throws IOException, NullPointerException {
        System.out.println("In the method 1111 ...");
        srbFile = indexDir;
        
        SRBHandler handler = new SRBHandler();
        srbFileSystem = handler.getSRBFileSystem();

        if (!srbFileSystem.isConnected()) 
            throw new IOException("Could not connect to filesystem");
        
        lockDir = new SRBFile( srbFile, "LOCK" );
        // Ensure that lockDir exists and is a directory.
        if (!lockDir.exists()) {
            if (!lockDir.mkdirs())
                throw new IOException("Cannot create directory: " + lockDir);
        } else if (!lockDir.isDirectory()) {
            throw new IOException("Found regular file where directory expected: " + lockDir);
        }        
    }     
   public SRBDirectory (String filename,  String resource, String host) 
        throws IOException, NullPointerException {
        System.out.println("In the method ...");
        SRBHandler handler = new SRBHandler();
        srbAccount = handler.getSRBAccount();
        srbAccount.setDefaultStorageResource(resource);
        srbAccount.setHost(host);
        srbFileSystem = new SRBFileSystem(srbAccount);

        

        if (!srbFileSystem.isConnected()) 
            throw new IOException("Could not connect to filesystem");
        srbFile = new SRBFile(srbFileSystem, filename);
        lockDir = new SRBFile( srbFile, "LOCK" );
        // Ensure that lockDir exists and is a directory.
        if (!lockDir.exists()) {
            if (!lockDir.mkdirs())
                throw new IOException("Cannot create directory: " + lockDir);
        } else if (!lockDir.isDirectory()) {
            throw new IOException("Found regular file where directory expected: " + lockDir);
        }        
    }
  
    public SRBDirectory (String indexDir) 
        throws IOException, NullPointerException {
        
        this(new SRBFile(new SRBHandler().getSRBFileSystem(), indexDir));
    }

    public SRBDirectory (URI uri) 
        throws IOException, NullPointerException {
        
        this(new SRBFile(uri));
    }    
    
    /** Returns an array of strings, one for each file in the directory. */
    public String[] list()
       throws IOException {

        String[] files = srbFile.list();
        if (files == null) return null;

        String[] result = new String[files.length];
        for (int i = 0; i < files.length; i++) {
            result[i] = files[i];
        }
        return result;        
    }

  /** Returns true iff a file with the given name exists. */
  public boolean fileExists(String name)
       throws IOException {
      
        SRBFile file = new SRBFile( srbFile, name );
        if (file.exists()) {
            return true;
        } else {
            return false;
        }
        // or: return fs.exists(new File(directory, name));
    }

  /** Returns the time the named file was last modified. */
  public long fileModified(String name)
       throws IOException {
        SRBFile file = new SRBFile( srbFile, name );
        return file.lastModified();
  }

  /** Set the modified time of an existing file to now. */
  public void touchFile(String name)
       throws IOException {
        SRBFile file = new SRBFile( srbFile, name );
        file.setLastModified(System.currentTimeMillis());
  }

  /** Removes an existing file in the directory. */
  public void deleteFile(String name)
       throws IOException {
      
        SRBFile file = new SRBFile( srbFile, name );
        if (!file.delete())
            throw new IOException("Cannot delete " + file.toString());
  }

  /** Renames an existing file in the directory.
    If a file already exists with the new name, then it is replaced.
    This replacement should be atomic. */
  public void renameFile(String from, String to)
       throws IOException {
      
      SRBFile old = new SRBFile( srbFile, from );
      SRBFile nu = new SRBFile( srbFile, to );

        /* This is not atomic.  If the program crashes between the call to
           delete() and the call to renameTo() then we're screwed, but I've
           been unable to figure out how else to do this... */

        if (nu.exists())
            if (!nu.delete())
                throw new IOException("Cannot delete " + nu);

        // Rename the old file to the new one. Unfortunately, the renameTo()
        // method does not work reliably under some JVMs.  Therefore, if the
        // rename fails, we manually rename by copying the old file to the new one
        if (!old.renameTo(nu)) {
            java.io.InputStream in = null;
            java.io.OutputStream out = null;
            try {
                in = new SRBFileInputStream(old);
                out = new SRBFileOutputStream(nu);
                // see if the buffer needs to be initialized. Initialization is
                // only done on-demand since many VM's will never run into the renameTo
                // bug and hence shouldn't waste 1K of mem for no reason.
                if (buffer == null) {
                  buffer = new byte[1024];
                }
                int len;
                while ((len = in.read(buffer)) >= 0) {
                  out.write(buffer, 0, len);
                }

                // delete the old file.
                old.delete();
            } catch (IOException ioe) {
                throw new IOException("Cannot rename " + old + " to " + nu);
            } finally {
                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException e) {
                        throw new RuntimeException("Cannot close input stream: " + e.toString());
                    }
                }
                if (out != null) {
                    try {
                        out.close();
                    } catch (IOException e) {
                        throw new RuntimeException("Cannot close output stream: " + e.toString());
                    }
                }
            }
        }
    }

    /** Returns the length of a file in the directory. */
    public long fileLength(String name)
        throws IOException {
        
    SRBFile file = new SRBFile( srbFile, name );
    return file.length();
  }

    /** Creates a new, empty file in the directory with the given name.
        Returns a stream writing this file. */
    public IndexOutput createOutput(String name) 
            throws IOException {
        
        SRBFile file = new SRBFile( srbFile, name );
        
        if (file.exists() && !file.delete())          // delete existing, if any
          throw new IOException("Cannot overwrite: " + file);

        return new SRBIndexOutput(file);
  }
    
    class SRBIndexOutput extends BufferedIndexOutput {
        SRBRandomAccessFile file = null;

        public SRBIndexOutput(SRBFile path) throws IOException {
            file = new SRBRandomAccessFile(path, "rw");
        }

        /** output methods: */
        public void flushBuffer(byte[] b, int size) throws IOException {
            file.write(b, 0, size);
        }
        
        public void close() throws IOException {
            super.close();
            file.close();
        }

        /** Random-access methods */
        public void seek(long pos) throws IOException {
            super.seek(pos);
            file.seek(pos);
        }
        
        public long length() throws IOException {
            return file.length();
        }

        protected void finalize() throws IOException {
            file.close();          // close the file
        }

    }    

  /** Returns a stream reading an existing file. */
  public IndexInput openInput(String name)
    throws IOException {
      return new SRBIndexInput(new SRBFile( srbFile, name )); 
  }

  /** Constructs a Lock with the specified name.  Locks are implemented
   * with {@link SRBFile#createNewFile()}.
   *
   * @param name the name of the lock file
   * @return an instance of <code>Lock</code> holding the lock
   */
  public Lock makeLock(String name) {
    StringBuffer buf = getLockPrefix();
    buf.append("-");
    buf.append(name);

    // create a lock file
    final SRBFile lockFile = new SRBFile(lockDir, buf.toString());

    return new Lock() {
      public boolean obtain() throws IOException {

        if (!lockDir.exists()) {
          if (!lockDir.mkdirs()) {
            throw new IOException("Cannot create lock directory: " + lockDir);
          }
        }

        return lockFile.createNewFile();
      }
      public void release() {
        lockFile.delete();
      }
      public boolean isLocked() {
        return lockFile.exists();
      }

      public String toString() {
        return "Lock@" + lockFile;
      }
    };
  }
  
  private StringBuffer getLockPrefix() {
    String dirName;                               // name to be hashed
    try {
      dirName = srbFile.getCanonicalPath();
    } catch (IOException e) {
      throw new RuntimeException(e.toString());
    }

    byte digest[];
    synchronized (DIGESTER) {
      digest = DIGESTER.digest(dirName.getBytes());
    }
    StringBuffer buf = new StringBuffer();
    buf.append("lucene-");
    for (int i = 0; i < digest.length; i++) {
      int b = digest[i];
      buf.append(HEX_DIGITS[(b >> 4) & 0xf]);
      buf.append(HEX_DIGITS[b & 0xf]);
    }

    return buf;
  }

  /** Closes the store. */
  public void close()
       throws IOException {
      srbFileSystem.close();
  }

class SRBIndexInput extends BufferedIndexInput {

  private class Descriptor extends SRBRandomAccessFile {
    public long position;
    public Descriptor(SRBFile file, String mode) throws IOException {
      super(file, mode);
    }
  }

  private Descriptor file = null;
  boolean isClone;
  private long length;

  public SRBIndexInput(SRBFile path) throws IOException {
    file = new Descriptor(path, "r");
    length = file.length();
  }

  /** IndexInput methods */
  protected void readInternal(byte[] b, int offset, int len)
       throws IOException {
    synchronized (file) {
      long position = getFilePointer();
      if (position != file.position) {
        file.seek(position);
        file.position = position;
      }
      int total = 0;
      do {
        int i = file.read(b, offset+total, len-total);
        if (i == -1)
          throw new IOException("read past EOF");
        file.position += i;
        total += i;
      } while (total < len);
    }
  }

  public void close() throws IOException {
    if (!isClone)
      file.close();
  }

  protected void seekInternal(long position) {
  }

  public long length() {
    return length;
  }

  protected void finalize() throws IOException {
    close();            // close the file
  }

  public Object clone() {
    FSIndexInput clone = (FSIndexInput)super.clone();
    clone.isClone = true;
    return clone;
  }

}  
  
}
